import tensorflow as tf


def clean_acc_sparse(y_true, y_pred):
    labels = tf.cast(y_true, 'int32')[:,0]

    return tf.reduce_mean(tf.cast(
        tf.equal(labels, tf.cast(tf.argmax(y_pred[:, :-1], axis=1), 'int32')),
        'float32'))

def clean_acc_cat(y_true, y_pred):
    labels = tf.argmax(y_true, axis=1, output_type='int32')[:,None]

    return clean_acc_sparse(labels, y_pred)

def clean_acc(y_true, y_pred):
    return tf.case([
        (tf.equal(tf.shape(y_true)[1], 1), 
            lambda: clean_acc_sparse(y_true, y_pred))], 
        default=lambda: clean_acc_cat(y_true, y_pred))


def vra_sparse(y_true, y_pred):
    labels = tf.cast(y_true, 'int32')[:,0]

    return tf.reduce_mean(tf.cast(
        tf.equal(labels, tf.cast(tf.argmax(y_pred, axis=1), 'int32')),
        'float32'))

def vra_cat(y_true, y_pred):
    labels = tf.argmax(y_true, axis=1, output_type='int32')[:,None]

    return vra_sparse(labels, y_pred)

def vra(y_true, y_pred):
    return tf.case([
        (tf.equal(tf.shape(y_true)[1], 1), 
            lambda: vra_sparse(y_true, y_pred))], 
        default=lambda: vra_cat(y_true, y_pred))


def rejection_rate(y_true, y_pred):
    bot_index = tf.cast(tf.shape(y_pred)[1] - 1, 'int64')

    return tf.reduce_mean(tf.cast(
        tf.argmax(y_pred, axis=1) == bot_index, 'float32'))



import tensorflow as tf

from tensorflow.keras.models import Model
from tensorflow.keras.layers import Activation
from tensorflow.keras.layers import Concatenate

import gloro

from gloro.models import GloroNet
from gloro.relaxations.affinity_sets import AffinitySet
from gloro.relaxations.layers import LipschitzMarginAffinity
from gloro.relaxations.layers import LipschitzMarginRtk
from gloro.training.callbacks import UpdatePowerIterates
from gloro.training.losses import get as get_loss


class RtkGloroNet(GloroNet):
    def __init__(
        self, 
        inputs=None, 
        outputs=None, 
        epsilon=None, 
        k=None,
        num_iterations=5,
        model=None, 
        _skip_init=False,
        **kwargs,
    ):
        if _skip_init:
            # We have to do this because keras requires __init__ to be called on
            # *all* subclasses.
            super().__init__(inputs, outputs, _skip_init=True, **kwargs)
            return

        if epsilon is None:
            raise ValueError('`epsilon` is required')

        if k is None:
            raise ValueError('`k` is required')
        
        if model is None and (inputs is None or outputs is None):
            raise ValueError(
                'must specify either `inputs` and `outputs` or `model`')
            
        if model is not None and inputs is not None and outputs is not None:
            raise ValueError(
                'cannot specify both `inputs` and `outputs` and `model`')

        if inputs is None or outputs is None:
            inputs = model.inputs
            outputs = model.outputs
            
        else:
            model = Model(inputs, outputs)

        if not isinstance(outputs, (list, tuple)):
            outputs = [outputs]
            
        if len(outputs) > 1:
            raise ValueError(
                f'only models with a single output are supported, but got '
                f'{len(outputs)} outputs')

        margin_layer = LipschitzMarginRtk(
            epsilon, k, model, num_iterations=num_iterations)

        margin = margin_layer(*outputs)

        margin, k_guarantee, loosest_k = margin

        margin = [
            Activation('linear', name='pred')(margin),
            Activation('linear', name='guarantee')(k_guarantee),
            Concatenate(name='pred_top_k')(
                [margin, loosest_k])
        ]
        
        super().__init__(inputs, margin, None, _skip_init=True, **kwargs)

        self._inputs = inputs
        self._output = outputs[0]
        self._margin_layer = margin_layer
        self._f = GloroNet._ModelContainer(model)

    @property
    def k(self):
        return self._margin_layer.k

    @k.setter
    def k(self, new_k):
        self._margin_layer.k = new_k

    def predict_with_guarantee(self, *args, **kwargs):
        return super().predict(*args, **kwargs)[0:2]

    def predict_with_loosest_guarantee(self, *args, **kwargs):
        preds, guarantees = super().predict(*args, **kwargs)[0:2]

        guarantees = 1 + tf.argmax(
            tf.cast(guarantees, 'float32') +
                # Adding this breaks ties in favor of larger k.
                tf.cast(tf.range(self.k), 'float32') * 1e-7,
            axis=1)

        # Guarantees are only valid on robust points. Otherwise indicate that
        # no guarantee was achieved with -1.
        guarantees = tf.where(
            tf.equal(tf.argmax(preds, axis=1), preds.shape[1] - 1),
            -1,
            guarantees)

        return [preds, guarantees.numpy()]

    def predict_with_strictest_guarantee(self, *args, **kwargs):
        preds, guarantees = super().predict(*args, **kwargs)[0:2]

        guarantees = 1 + tf.argmax(guarantees, axis=1).numpy()

        # Guarantees are only valid on robust points. Otherwise indicate that
        # no guarantee was achieved with -1.
        guarantees = tf.where(
            tf.equal(tf.argmax(preds, axis=1), preds.shape[1] - 1),
            -1,
            guarantees)

        return [preds, guarantees.numpy()]


    # Overriding `keras.Model` functions.

    def predict(self, *args, **kwargs):
        return super().predict(*args, **kwargs)[0]

    def compile(self, **kwargs):
        if 'loss' in kwargs:
            if isinstance(kwargs['loss'], str):
                kwargs['loss'] = {'pred': get_loss(kwargs['loss'])}
            else:
                kwargs['loss'] = {'pred': kwargs['loss']}

        if 'metrics' in kwargs:
            metrics = kwargs['metrics']
            new_metrics = []

            includes_rtk_vra = False
            rtk_vra = None
            for metric in metrics:
                if metric in gloro.constants.GLORO_CUSTOM_OBJECTS:
                    metric = gloro.constants.GLORO_CUSTOM_OBJECTS[metric]

                if (
                    hasattr(metric, '__name__') and (
                        metric.__name__.startswith('rtk_vra') or
                        metric.__name__.startswith('affinity_vra'))
                ):
                    includes_rtk_vra = True
                    rtk_vra = metric

                else:
                    new_metrics.append(metric)

            kwargs['metrics'] = {'pred': new_metrics}

            if includes_rtk_vra:
                kwargs['metrics']['pred_top_k'] = rtk_vra

        Model.compile(self, **kwargs)

    def get_config(self):
        return {
            'epsilon': float(self.epsilon),
            'k': int(self.k),
            'num_iterations': int(self.num_iterations),
        }


class AffinityGloroNet(RtkGloroNet):
    def __init__(
        self,
        inputs=None,
        outputs=None,
        epsilon=None,
        affinity_sets=None,
        num_iterations=5,
        model=None,
        _skip_init=False,
        **kwargs,
    ):
        if _skip_init:
            # We have to do this because keras requires __init__ to be called on
            # *all* subclasses.
            super().__init__(inputs, outputs, **kwargs)
            return

        if epsilon is None:
            raise ValueError('`epsilon` is required')

        if affinity_sets is None:
            raise ValueError('`affinity_sets` is required')

        if model is None and (inputs is None or outputs is None):
            raise ValueError(
                'must specify either `inputs` and `outputs` or `model`')

        if model is not None and inputs is not None and outputs is not None:
            raise ValueError(
                'cannot specify both `inputs` and `outputs` and `model`')

        if inputs is None or outputs is None:
            inputs = model.inputs
            outputs = model.outputs

        else:
            model = Model(inputs, outputs)

        if not isinstance(outputs, (list, tuple)):
            outputs = [outputs]

        if len(outputs) > 1:
            raise ValueError(
                f'only models with a single output are supported, but got '
                f'{len(outputs)} outputs')

        if isinstance(affinity_sets, AffinitySet):
            pass

        elif isinstance(affinity_sets, (list, tuple)):
            affinity_sets = AffinitySet.from_class_indices(affinity_sets)

        elif isinstance(affinity_sets, np.ndarray):
            affinity_sets = AffinitySet.from_one_hot(affinity_sets)

        elif isinstance(affinity_sets, str):
            affinity_sets = AffinitySet.from_string(affinity_sets)

        else:
            raise ValueError(
                f'unexpected type for `affinity_sets`: {type(affinity_sets)}')

        if affinity_sets.mask.shape[1] != model.output_shape[1]:
            raise ValueError(
                f'affinity sets defined for a number of classes that does not '
                f'match the output shape of the model: '
                f'{affinity_sets.mask.shape[1]} vs {model.output_shape[1]}')

        k = affinity_sets.max_set_size

        margin_layer = LipschitzMarginAffinity(
            epsilon, k, affinity_sets, model, num_iterations=num_iterations)

        margin = margin_layer(*outputs)

        margin, k_guarantee, loosest_k = margin

        margin = [
            Activation('linear', name='pred')(margin),
            Activation('linear', name='guarantee')(k_guarantee),
            Concatenate(name='pred_top_k')(
                [margin, loosest_k])
        ]

        super().__init__(inputs, margin, None, None, _skip_init=True, **kwargs)

        self._inputs = inputs
        self._output = outputs[0]
        self._margin_layer = margin_layer
        self._f = GloroNet._ModelContainer(model)

    @property
    def affinity_sets(self):
        return self._margin_layer.affinity_sets

    def get_config(self):
        return {
            'epsilon': float(self.epsilon),
            'affinity_sets': str(self.affinity_sets),
            'num_iterations': int(self.num_iterations),
        }



from tensorflow.keras.initializers import GlorotUniform
from tensorflow.keras.initializers import Orthogonal
from tensorflow.keras.layers import Activation
from tensorflow.keras.layers import AveragePooling2D
from tensorflow.keras.layers import Conv2D
from tensorflow.keras.layers import Dense
from tensorflow.keras.layers import Dropout
from tensorflow.keras.layers import Flatten
from tensorflow.keras.layers import Input

from gloro.layers import InvertibleDownsampling
from gloro.layers import MinMax
from gloro.layers import ResnetBlock


def _add_pool(z, pooling_type, activation=None, initialization='orthogonal'):
    if pooling_type == 'avg':
        return AveragePooling2D()(z)

    elif pooling_type == 'conv':
        channels = z.shape[-1]

        z = Conv2D(
            channels, 
            4, 
            strides=2, 
            padding='same', 
            kernel_initializer=initialization)(z)

        return _add_activation(z, activation)

    elif pooling_type == 'invertible':
        return InvertibleDownsampling()(z)

    else:
        raise ValueError(f'unknown pooling type: {pooling_type}')

def _add_activation(z, activation_type='relu'):
    if activation_type == 'relu':
        return Activation('relu')(z)

    elif activation_type == 'elu':
        return Activation('elu')(z)

    elif activation_type == 'softplus':
        return Activation('softplus')(z)

    elif activation_type == 'minmax':
        return MinMax()(z)

    else:
        raise ValueError(f'unknown activation type: {activation_type}')


def cnn_simple(
    input_shape, 
    num_classes, 
    pooling='conv', 
    activation='relu',
    initialization='orthogonal',
):
    x = Input(input_shape)
    z = Conv2D(32, 3, padding='same', kernel_initializer=initialization)(x)
    z = _add_activation(z, activation)
    z = _add_pool(z, pooling, activation, initialization)

    z = Conv2D(64, 3, padding='same', kernel_initializer=initialization)(z)
    z = _add_activation(z, activation)
    z = _add_pool(z, pooling, activation, initialization)

    z = Flatten()(z)
    z = Dense(256, kernel_initializer=initialization)(z)
    z = _add_activation(z, activation)

    z = Dense(256, kernel_initializer=initialization)(z)
    z = _add_activation(z, activation)

    y = Dense(num_classes, kernel_initializer=initialization)(z)

    return x, y

def minmax_cnn_simple(
    input_shape, 
    num_classes, 
    pooling='invertible', 
    initialization='orthogonal',
    normalize_lc=False,
):
    return cnn_simple(
        input_shape, num_classes, 
        pooling=pooling, 
        activation='minmax',
        initialization=initialization)


def cnn_2C2F(
    input_shape, 
    num_classes, 
    activation='relu',
    initialization='orthogonal',
):
    x = Input(input_shape)
    z = Conv2D(
        16, 4, strides=2, padding='same', kernel_initializer=initialization)(x)
    z = _add_activation(z, activation)

    z = Conv2D(
        32, 4, strides=2, padding='same', kernel_initializer=initialization)(z)
    z = _add_activation(z, activation)

    z = Flatten()(z)
    z = Dense(100, kernel_initializer=initialization)(z)
    z = _add_activation(z, activation)
    y = Dense(num_classes, kernel_initializer=initialization)(z)

    return x, y

def minmax_cnn_2C2F(
    input_shape, 
    num_classes, 
    pooling='conv', 
    initialization='orthogonal',
):
    return cnn_2C2F(
        input_shape, num_classes, 
        pooling=pooling, 
        activation='minmax',
        initialization=initialization)


def cnn_4C3F(
    input_shape, 
    num_classes, 
    pooling='conv', 
    activation='relu',
    initialization='orthogonal',
):
    x = Input(input_shape)
    z = Conv2D(32, 3, padding='same', kernel_initializer=initialization)(x)
    z = _add_activation(z, activation)
    z = _add_pool(z, pooling, activation, initialization)

    z = Conv2D(64, 3, padding='same', kernel_initializer=initialization)(z)
    z = _add_activation(z, activation)
    z = _add_pool(z, pooling, activation, initialization)

    z = Flatten()(z)
    z = Dense(512, kernel_initializer=initialization)(z)
    z = _add_activation(z, activation)
    z = Dense(512, kernel_initializer=initialization)(z)
    z = _add_activation(z, activation)
    y = Dense(num_classes, kernel_initializer=initialization)(z)

    return x, y

def minmax_cnn_4C3F(
    input_shape, 
    num_classes, 
    pooling='conv', 
    initialization='orthogonal',
):
    return cnn_4C3F(
        input_shape, num_classes, 
        pooling=pooling, 
        activation='minmax',
        initialization=initialization)


def cnn_6C2F(
    input_shape, 
    num_classes, 
    pooling='conv', 
    activation='relu',
    initialization='orthogonal',
):
    x = Input(input_shape)
    z = Conv2D(32, 3, padding='same', kernel_initializer=initialization)(x)
    z = _add_activation(z, activation)
    z = Conv2D(32, 3, padding='same', kernel_initializer=initialization)(z)
    z = _add_activation(z, activation)
    z = _add_pool(z, pooling, activation, initialization)

    z = Conv2D(64, 3, padding='same', kernel_initializer=initialization)(z)
    z = _add_activation(z, activation)
    z = Conv2D(64, 3, padding='same', kernel_initializer=initialization)(z)
    z = _add_activation(z, activation)
    z = _add_pool(z, pooling, activation, initialization)

    z = Flatten()(z)
    z = Dense(512, kernel_initializer=initialization)(z)
    z = _add_activation(z, activation)
    y = Dense(num_classes, kernel_initializer=initialization)(z)

    return x, y

def minmax_cnn_6C2F(
    input_shape, 
    num_classes, 
    pooling='conv', 
    initialization='orthogonal',
):
    return cnn_6C2F(
        input_shape, num_classes, 
        pooling=pooling, 
        activation='minmax',
        initialization=initialization)


def cnn_8C2F(
    input_shape, 
    num_classes, 
    pooling='conv', 
    activation='relu',
    initialization='orthogonal',
):
    x = Input(input_shape)
    z = Conv2D(64, 3, padding='same', kernel_initializer=initialization)(x)
    z = _add_activation(z, activation)
    z = Conv2D(64, 3, padding='same', kernel_initializer=initialization)(z)
    z = _add_activation(z, activation)
    z = _add_pool(z, pooling, activation, initialization)

    z = Conv2D(128, 3, padding='same', kernel_initializer=initialization)(z)
    z = _add_activation(z, activation)
    z = Conv2D(128, 3, padding='same', kernel_initializer=initialization)(z)
    z = _add_activation(z, activation)
    z = _add_pool(z, pooling, activation, initialization)

    z = Conv2D(256, 3, padding='same', kernel_initializer=initialization)(z)
    z = _add_activation(z, activation)
    z = _add_pool(z, pooling, activation, initialization)

    z = Flatten()(z)
    z = Dense(256, kernel_initializer=initialization)(z)
    z = _add_activation(z, activation)
    y = Dense(num_classes, kernel_initializer=initialization)(z)

    return x, y

def minmax_cnn_8C2F(
    input_shape, 
    num_classes, 
    pooling='conv', 
    initialization='orthogonal',
):
    return cnn_8C2F(
        input_shape, num_classes, 
        pooling=pooling, 
        activation='minmax',
        initialization=initialization)


def alexnet(
    input_shape,
    num_classes,
    pooling='avg',
    activation='relu',
    initialization='orthogonal',
    dropout=False,
):
    x = Input(input_shape)

    z = Conv2D(
        96,
        11,
        padding='same',
        strides=4,
        kernel_initializer=initialization)(x)
    z = _add_activation(z, activation)
    z = _add_pool(z, pooling, activation, initialization)

    z = Conv2D(256, 5, padding='same', kernel_initializer=initialization)(z)
    z = Activation('relu')(z)
    z = _add_pool(z, pooling, activation, initialization)

    z = Conv2D(384, 3, padding='same', kernel_initializer=initialization)(z)
    z = _add_activation(z, activation)
    z = Conv2D(384, 3, padding='same', kernel_initializer=initialization)(z)
    z = _add_activation(z, activation)
    z = Conv2D(384, 3, padding='same', kernel_initializer=initialization)(z)
    z = _add_activation(z, activation)
    z = _add_pool(z, pooling, activation, initialization)

    z = Flatten()(z)
    if dropout:
        z = Dropout(0.5)(z)
    z = Dense(4096, kernel_initializer=initialization)(z)
    z = _add_activation(z, activation)
    if dropout:
        z = Dropout(0.5)(z)
    z = Dense(4096, kernel_initializer=initialization)(z)
    z = _add_activation(z, activation)
    
    y = Dense(num_classes, kernel_initializer=initialization)(z)

    return x, y

def minmax_alexnet(
    input_shape,
    num_classes,
    pooling='invertible',
    initialization='orthogonal',
    dropout=False,
):
    return alexnet(
        input_shape, num_classes, 
        pooling=pooling,
        activation='minmax',
        initialization=initialization,
        dropout=dropout)


def vgg16(
    input_shape,
    num_classes,
    pooling='avg',
    activation='relu',
    initialization='orthogonal',
):
    x = Input(input_shape)

    z = Conv2D(64, 3, padding='same', kernel_initializer=initialization)(x)
    z = _add_activation(z, activation)
    z = Conv2D(64, 3, padding='same', kernel_initializer=initialization)(z)
    z = _add_activation(z, activation)
    z = _add_pool(z, pooling, activation, initialization)

    z = Conv2D(128, 3, padding='same', kernel_initializer=initialization)(z)
    z = _add_activation(z, activation)
    z = Conv2D(128, 3, padding='same', kernel_initializer=initialization)(z)
    z = _add_activation(z, activation)
    z = _add_pool(z, pooling, activation, initialization)

    z = Conv2D(256, 3, padding='same', kernel_initializer=initialization)(z)
    z = _add_activation(z, activation)
    z = Conv2D(256, 3, padding='same', kernel_initializer=initialization)(z)
    z = _add_activation(z, activation)
    z = Conv2D(256, 3, padding='same', kernel_initializer=initialization)(z)
    z = _add_activation(z, activation)
    z = _add_pool(z, pooling, activation, initialization)

    z = Conv2D(512, 3, padding='same', kernel_initializer=initialization)(z)
    z = _add_activation(z, activation)
    z = Conv2D(512, 3, padding='same', kernel_initializer=initialization)(z)
    z = _add_activation(z, activation)
    z = Conv2D(512, 3, padding='same', kernel_initializer=initialization)(z)
    z = _add_activation(z, activation)
    z = _add_pool(z, pooling, activation, initialization)

    z = Conv2D(512, 3, padding='same', kernel_initializer=initialization)(z)
    z = _add_activation(z, activation)
    z = Conv2D(512, 3, padding='same', kernel_initializer=initialization)(z)
    z = _add_activation(z, activation)
    z = Conv2D(512, 3, padding='same', kernel_initializer=initialization)(z)
    z = _add_activation(z, activation)
    z = _add_pool(z, pooling, activation, initialization)

    z = Flatten()(z)
    z = Dense(4096, kernel_initializer=initialization)(z)
    z = _add_activation(z, activation)
    z = Dense(4096, kernel_initializer=initialization)(z)
    z = _add_activation(z, activation)

    y = Dense(num_classes, kernel_initializer=initialization)(z)

    return x, y

def minmax_vgg16(
        input_shape,
        num_classes,
        pooling='invertible',
        initialization='orthogonal'):

    return vgg16(
        input_shape, num_classes, 
        pooling=pooling,
        activation='minmax',
        initialization=initialization)


def resnet_tiny(
    input_shape,
    num_classes,
    pooling='avg',
    activation='relu',
    initialization='orthogonal',
    fixup_residual_scaling=False,
    identity_skip=False,
):
    x = Input(input_shape)

    z = Conv2D(64, (5, 5), strides=(1, 1), padding='same')(x)
    z = _add_activation(z, activation)
    z = _add_pool(z, pooling, activation, initialization)

    z = ResnetBlock(
        filters=(128, 128, 128),
        kernel_sizes=(3, 3, 1),
        stride1=2,
        activation=activation,
        use_invertible_downsample=pooling == 'invertible',
        kernel_initializer=initialization,
        use_fixup_weight_and_bias=fixup_residual_scaling,
        identity_skip=identity_skip)(z)
    z = ResnetBlock(
        filters=(256, 256, 256),
        kernel_sizes=(3, 3, 1),
        stride1=2,
        activation=activation,
        use_invertible_downsample=pooling == 'invertible',
        kernel_initializer=initialization,
        use_fixup_weight_and_bias=fixup_residual_scaling,
        identity_skip=identity_skip)(z)
    z = ResnetBlock(
        filters=(512, 512, 512),
        kernel_sizes=(3, 3, 1),
        stride1=2,
        activation=activation,
        use_invertible_downsample=pooling == 'invertible',
        kernel_initializer=initialization,
        use_fixup_weight_and_bias=fixup_residual_scaling,
        identity_skip=identity_skip)(z)

    z = _add_pool(z, pooling, activation, initialization)

    z = Flatten()(z)
    z = Dense(512, kernel_initializer=initialization)(z)
    z = _add_activation(z, activation)

    y = Dense(num_classes, kernel_initializer=initialization)(z)

    return x, y

def minmax_resnet_tiny(
    input_shape,
    num_classes,
    pooling='invertible',
    initialization='orthogonal',
    fixup_residual_scaling=False,
    identity_skip=False,
):
    return resnet_tiny(
        input_shape, num_classes, 
        pooling=pooling, 
        activation='minmax',
        initialization=initialization,
        fixup_residual_scaling=fixup_residual_scaling,
        identity_skip=identity_skip)



import numpy as np
import tensorflow as tf

from tensorflow.python.framework import ops

import gloro


def add_extra_column(y):
    if isinstance(y, np.ndarray):
        return add_extra_column_np(y)
    else:
        return add_extra_column_tf(y)

def add_extra_column_tf(y):
    return tf.concat((y, tf.zeros((tf.shape(y)[0], 1))), axis=1)

def add_extra_column_np(y):
    return np.concatenate((y, np.zeros((y.shape[0], 1))), axis=1)


def set_value(x, value):
    value = np.asarray(value, dtype=x.dtype.base_dtype.name)

    with ops.init_scope():
        x.assign(value)


def batch_set_value(tuples):
    with ops.init_scope():
        for x, value in tuples:
            x.assign(np.asarray(value, dtype=x.dtype.base_dtype.name))


def get_value(x):
    return x.numpy()


def l2_normalize(x):
    return x / (tf.sqrt(tf.reduce_sum(x**2.)) + gloro.constants.EPS)


def print_if_verbose(verbose):
    if verbose:
        return lambda s: print(s)

    else:
        return lambda s: None



import numpy as np
import tensorflow as tf
import tensorflow.keras.backend as K

from tensorflow.keras.layers import Layer

from gloro.lipschitz_computation import LipschitzComputer
from gloro.utils import batch_set_value
from gloro.utils import get_value
from gloro.utils import set_value


class LipschitzMargin(Layer):
    def __init__(self, epsilon, model, num_iterations=5, norm='l2', **kwargs):
        super().__init__(**kwargs)

        self._epsilon = tf.Variable(
            epsilon, dtype=K.floatx(), name='epsilon', trainable=False)
        self._num_iterations = tf.Variable(
            num_iterations, name='num_iterations', trainable=False)

        # Check norm compatibility.
        if norm != 'l2':
            # For now we only support the l2 norm.
            raise ValueError(f'{norm} norm not supported')

        self._norm = norm

        self._layer_computers = LipschitzComputer.for_model(
            model, self._num_iterations)
        self._W = model.layers[-1].kernel

        self._lc_frozen = tf.Variable(False, name='lc_frozen', trainable=False)
        self._hardcoded_lc = tf.Variable(
            1., name='hardcoded_lc', trainable=False)

    @property
    def epsilon(self):
        return get_value(self._epsilon)

    @epsilon.setter
    def epsilon(self, new_epsilon):
        set_value(self._epsilon, new_epsilon)

    @property
    def num_iterations(self):
        return get_value(self._num_iterations)

    @num_iterations.setter
    def num_iterations(self, new_num_iterations):
        set_value(self._num_iterations, new_num_iterations)

    @property
    def norm(self):
        return self._norm

    @property
    def frozen(self):
        return get_value(self._lc_frozen)

    def freeze(self):
        set_value(self._hardcoded_lc, self._lc())
        set_value(self._lc_frozen, True)

    
    def build(self, input_shape):

        self._lc = lambda: (
            self._hardcoded_lc if self._lc_frozen else
            LipschitzComputer.global_lipschitz_bound(self._layer_computers))

        # This is the L2 norm of the Lipschitz constant of the penultimate layer
        # of the network multiplied by the weights of the penulimate layer.
        self._kW = lambda: self._lc() * self._W

    def _get_yj_j_Kij(self, y):

        kW = self._kW()

        j = tf.argmax(y, axis=1)
        y_j = tf.reduce_max(y, axis=1, keepdims=True)
        where_not_j = tf.not_equal(y, y_j)

        # Get the weight column of the predicted class.
        kW_j = tf.gather(tf.transpose(kW), j)

        # Get weights that predict the value y_j - y_i for all i != j.
        kW_ij = kW_j[:,:,None] - kW[None]
        
        # We do this instead of `tf.linalg.norm(W_d)` because of an apparent bug
        # in `tf.linalg.norm` that leads to NaN values.
        K_ij = tf.sqrt(tf.reduce_sum(kW_ij * kW_ij, axis=1))

        return y_j, j, K_ij
    
    def lipschitz_constant(self, y):

        y_j, j, K_ij = self._get_yj_j_Kij(y)

        return tf.where(
            tf.equal(y, y_j), 
            tf.zeros_like(K_ij) - 1., 
            K_ij)

    def call(self, y):
        
        y_j, j, K_ij = self._get_yj_j_Kij(y)

        y_bot_i = y + self._epsilon * K_ij

        # `y_bot_i` will be zero at the position of class j. However, we don't 
        # want to consider this class, so we replace the zero with negative
        # infinity so that when we find the maximum component for `y_bot_i` we 
        # don't get zero as a result of all of the components we care aobut 
        # being negative.
        y_bot_i = tf.where(
            tf.equal(y, y_j), 
            -np.infty + tf.zeros_like(y_bot_i), 
            y_bot_i)

        y_bot = tf.reduce_max(y_bot_i, axis=1, keepdims=True)

        return tf.concat([y, y_bot], axis=1)

    def refresh_iterates(
        self, 
        converge=True, 
        convergence_threshold=1e-4, 
        max_tries=1000, 
        batch_size=None,
        verbose=False,
    ):
        if self._lc_frozen:
            return self

        batch_set_value([
            (
                layer.iterate, 
                tf.random.truncated_normal(layer.iterate.shape))
            for layer in self._layer_computers
            if hasattr(layer, 'iterate')
        ])

        if converge:
            if batch_size:
                old_num_iterations = self.num_iterations
                self.num_iterations = batch_size

            i = 0
            k_prev = None
            k = self._lc()

            while i < max_tries and (
                k_prev is None or abs(k - k_prev) > convergence_threshold
            ):
                k_prev = k
                k = self._lc()
                i += 1

            if verbose:
                print(
                    f'power method converged in {(i+1)*self.num_iterations} '
                    f'iterations to {get_value(k):.3f}')

            if batch_size:
                self.num_iterations = old_num_iterations

        return self

    def update_iterates(self):
        if self._lc_frozen:
            return self

        self._lc()

        return self



import tensorflow as tf

from tensorflow.keras.layers import Activation
from tensorflow.keras.layers import Add
from tensorflow.keras.layers import Conv2D
from tensorflow.keras.layers import Flatten
from tensorflow.keras.layers import Layer

import gloro

from gloro.utils import get_value
from gloro.utils import set_value


class MinMax(Layer):
    def __init__(self, absolute_value=False, **kwargs):
        super().__init__(**kwargs)

        self._do_abs = absolute_value
        self._flat_op = Flatten()
        
    def call(self, x):
        x_flat = self._flat_op(x)
        x_shape = tf.shape(x_flat)

        grouped_x = tf.reshape(
            x_flat, 
            tf.concat([x_shape[:-1], (-1, 2)], -1))

        min_x = tf.reduce_min(grouped_x, axis=-1, keepdims=True)
        max_x = tf.reduce_max(grouped_x, axis=-1, keepdims=True)

        sorted_x = tf.reshape(
            tf.concat([min_x, max_x], axis=-1), 
            tf.shape(x))

        if self._do_abs:
            return tf.math.abs(sorted_x)
        else:
            return sorted_x

    def get_config(self):
        config = {
            'absolute_value': self._do_abs,
        }
        base_config = super().get_config()

        return dict(list(base_config.items()) + list(config.items()))


class InvertibleDownsampling(Layer):
    def __init__(self, pool_size=2, **kwargs):
        super().__init__(**kwargs)

        self._pool_size = pool_size

    @property
    def pool_size(self):
        if isinstance(self._pool_size, (tuple, list)):
            return self._pool_size

        return (self._pool_size, self._pool_size)

    def call(self, inputs, **kwargs):
        return tf.concat(
            [
                # Expects channels last format.
                inputs[:, i :: self.pool_size[0], j :: self.pool_size[1], :] 
                for i in range(self.pool_size[0])
                for j in range(self.pool_size[1])
            ],
            axis=-1)

    def get_config(self):
        config = {
            'pool_size': self._pool_size,
        }
        base_config = super().get_config()

        return dict(list(base_config.items()) + list(config.items()))


class Scaling(Layer):
    def __init__(self, init=1., **kwargs):
        super().__init__(**kwargs)

        self._weight = tf.Variable(init, name='w', trainable=True)

    @property
    def w(self):
        return get_value(self._weight)

    @w.setter
    def w(self, new_w):
        set_value(self._weight, new_w)
        
    def call(self, x):
        return self._weight * x

    def get_config(self):
        config = {
            'init': self.w,
        }
        base_config = super().get_config()

        return dict(list(base_config.items()) + list(config.items()))


class Bias(Layer):
    def __init__(self, init=0., **kwargs):
        super().__init__(**kwargs)

        self._bias = tf.Variable(init, name='b', trainable=True)

    @property
    def b(self):
        return get_value(self._bias)

    @b.setter
    def b(self, new_b):
        set_value(self._bias, new_b)
        
    def call(self, x):
        return self._bias + x

    def get_config(self):
        config = {
            'init': self.b,
        }
        base_config = super().get_config()

        return dict(list(base_config.items()) + list(config.items()))


class ResnetBlock(object):

    _global_counter = 0

    identifier = 'ResnetBlock'
    join_identifier = 'join'
    skip_identifier = 'skip'

    def __init__(
        self, 
        filters, 
        kernel_sizes=(1,3,1),
        stride1=2,
        depth=None, 
        activation='relu',
        identity_skip=False,
        use_invertible_downsample=False,
        use_fixup_weight_and_bias=False,
        kernel_initializer='orthogonal',
    ):
        n_filters = len(filters) if isinstance(filters, (list, tuple)) else None
        n_kernels = (
            len(kernel_sizes) if isinstance(kernel_sizes, (list, tuple)) else 
            None)

        if (n_filters is not None and n_kernels is not None and 
                n_filters != n_kernels):

            raise ValueError(
                f'if `filters` and `kernel_sizes` are provided as lists, their '
                f'lengths must match, but got {n_filters} and {n_kernels}')

        if depth is None:
            if n_filters is None and n_kernels is None:
                raise ValueError(
                    'must either provide `depth` or give `filters`/'
                    '`kernel_sizes` as a list or tuple')

            depth = n_filters if n_filters is not None else n_kernels

        else:
            if n_filters is not None and n_filters != depth:
                raise ValueError(
                    'length of `filters` must match `depth` (if both given)')

            if n_kernels is not None and n_kernels != depth:
                raise ValueError(
                    'length of `kernel_sizes` must match `depth` (if both '
                    'given)')
        
        if not(stride1 in [None, 1]) and identity_skip:
            raise ValueError(
                'stride1 must be `None` or 1 when using `identity_skip`')
        
        self._filters = (
            [filters for _ in range(depth)] if n_filters is None else 
            filters)
        self._kernel_sizes = (
            [kernel_sizes for _ in range(depth)] if n_kernels is None else 
            kernel_sizes) 
        self._depth = depth
        self._stride1 = stride1

        self._identity_skip = identity_skip
        self._use_invertible_downsample = use_invertible_downsample
        self._use_fixup_weight_and_bias = use_fixup_weight_and_bias
        
        self._initializer = kernel_initializer

        if activation == 'minmax':
            self._activation = lambda name: MinMax(name=name)
        elif isinstance(activation, str):
            self._activation = lambda name: Activation(activation, name=name)
        else:
            self._activation = lambda name: activation(name=name)

        self._block_name = (
            f'{ResnetBlock.identifier}{ResnetBlock._global_counter}')
        ResnetBlock._global_counter += 1

    def __call__(self, input_tensor):
        i = 0
        out = input_tensor
        for filters_i, kernel_size_i in zip(self._filters, self._kernel_sizes):

            if self._use_fixup_weight_and_bias:
                next_layer = Bias(name=f'{self._block_name}_bias_pre{i}')
                next_layer._gloro_branch = gloro.constants.RESIDUAL_BRANCH

                out = next_layer(out)

            next_layer = Conv2D(
                filters_i, 
                kernel_size_i,
                strides=(
                    (self._stride1, self._stride1) 
                    if (i == 1 and (not self._use_invertible_downsample)) else 
                    (1, 1)),
                padding='same',
                use_bias=not self._use_fixup_weight_and_bias,
                kernel_initializer=self._initializer,
                name=f'{self._block_name}_conv{i}')
            next_layer._gloro_branch = gloro.constants.RESIDUAL_BRANCH

            out = next_layer(out)

            if i < self._depth - 1:
                if self._use_fixup_weight_and_bias:
                    next_layer = Bias(name=f'{self._block_name}_bias_post{i}')
                    next_layer._gloro_branch = gloro.constants.RESIDUAL_BRANCH

                    out = next_layer(out)

                next_layer = self._activation(
                    f'{self._block_name}_activation{i}')
                next_layer._gloro_branch = gloro.constants.RESIDUAL_BRANCH

                out = next_layer(out)

            elif self._use_fixup_weight_and_bias:
                next_layer = Scaling(name=f'{self._block_name}_scaling')
                next_layer._gloro_branch = gloro.constants.RESIDUAL_BRANCH

                out = next_layer(out)
            
            if i == 1 and self._stride1 > 1 and self._use_invertible_downsample:
                next_layer = InvertibleDownsampling(
                    self._stride1, 
                    name=f'{self._block_name}_downsampling')
                next_layer._gloro_branch = gloro.constants.RESIDUAL_BRANCH

                out = next_layer(out)

            i += 1

        if self._identity_skip:
            next_layer = Add(
                name=f'{self._block_name}_{ResnetBlock.join_identifier}')
            next_layer._gloro_branch = gloro.constants.MAIN_BRANCH

            out = next_layer([input_tensor, out])

        else:
            next_layer = Conv2D(
                self._filters[-1],
                1, 
                strides=(self._stride1, self._stride1),
                kernel_initializer=self._initializer,
                name=f'{self._block_name}_conv_{ResnetBlock.skip_identifier}')
            next_layer._gloro_branch = gloro.constants.SKIP_BRANCH

            conv_skip = next_layer(input_tensor)

            next_layer = Add(
                name=f'{self._block_name}_{ResnetBlock.join_identifier}')
            next_layer._gloro_branch = gloro.constants.MAIN_BRANCH

            out = next_layer([conv_skip, out])

        next_layer = self._activation(None)
        next_layer._gloro_branch = gloro.constants.MAIN_BRANCH

        return next_layer(out)



import tensorflow as tf

from tensorflow.keras.layers import Add
from tensorflow.keras.layers import AveragePooling2D

import gloro

from gloro.layers.network_layers import ResnetBlock
from gloro.utils import l2_normalize


class LipschitzComputer(object):

    def __init__(self, layer, *args, **kwargs):
        self._name = layer.name

        if hasattr(layer, '_gloro_branch'):
            self._branch = layer._gloro_branch

        elif layer.name.startswith(ResnetBlock.identifier):
            # TODO: this is a little less nice than reading a `_gloro_branch` 
            #   property, but it persists by default when the layers are saved,
            #   whereas we would need extra instrumentation to save the
            #   `_gloro_branch` property. Ultimately we should probably pick
            #   just one method (either name-based or property-based).
            if ResnetBlock.join_identifier in layer.name:
                self._branch = gloro.constants.MAIN_BRANCH

            elif ResnetBlock.skip_identifier in layer.name:
                self._branch = gloro.constants.SKIP_BRANCH

            else:
                self._branch = gloro.constants.RESIDUAL_BRANCH

        else:
            self._branch = gloro.constants.MAIN_BRANCH

    @property
    def name(self):
        return self._name

    @property
    def branch(self):
        return self._branch

    @staticmethod
    def for_layer(layer, num_iterations):
        if hasattr(layer, 'kernel'):
            if len(layer.kernel.shape) == 4:
                return ConvLayerComputer(layer, num_iterations)

            else:
                return DenseLayerComputer(layer, num_iterations)

        elif isinstance(layer, gloro.layers.Scaling):
            return ScalingLayerComputer(layer)

        elif isinstance(layer, Add):
            return JoinLayerComputer(layer)

        elif isinstance(layer, AveragePooling2D):
            return AveragePoolingComputer(layer)

        else:
            return LipschitzComputer(layer)

    @staticmethod
    def for_model(model, num_iterations, exclude_last_layer=True):
        layers = model.layers[:-1] if exclude_last_layer else model.layers

        return [
            LipschitzComputer.for_layer(layer, num_iterations)
            for layer in layers
        ]

    @staticmethod
    def global_lipschitz_bound(layer_computers):
        lc = {
            gloro.constants.MAIN_BRANCH: 1.,
            gloro.constants.RESIDUAL_BRANCH: 1.,
            gloro.constants.SKIP_BRANCH: 1.,
        }

        for layer in layer_computers:
            lc[layer.branch] *= layer.get_lipschitz_constant(lc=lc)

        return lc[gloro.constants.MAIN_BRANCH]

    def get_lipschitz_constant(self, **kwargs):
        return 1.


class DenseLayerComputer(LipschitzComputer):
    def __init__(self, layer, num_iterations):
        super().__init__(layer)

        self._W = layer.kernel
        self._iterate = tf.Variable(
            tf.random.truncated_normal((layer.kernel.shape[1], 1)),
            dtype='float32',
            trainable=False)

        self._while_cond = lambda i, _: i < num_iterations

    @property
    def W(self):
        return self._W

    @property
    def iterate(self):
        return self._iterate

    def get_lipschitz_constant(self, **kwargs):

        def body(i, x):
            x = l2_normalize(x)
            x_p = self.W @ x
            x = tf.transpose(self.W) @ x_p

            return i + 1, x

        _, x = tf.while_loop(
            self._while_cond, body, [tf.constant(0), self.iterate])

        # Update the power iterate.
        self.iterate.assign(x)

        return tf.sqrt(
            tf.reduce_sum((self.W @ x)**2.) / 
            (tf.reduce_sum(x**2.) + gloro.constants.EPS))


class ConvLayerComputer(LipschitzComputer):
    def __init__(self, layer, num_iterations):
        super().__init__(layer)

        self._W = layer.kernel
        self._strides = layer.strides
        self._padding = layer.padding.upper()
        self._iterate = tf.Variable(
            tf.random.truncated_normal((1, *layer.input_shape[1:])),
            dtype='float32',
            trainable=False)

        self._while_cond = lambda i, _: i < num_iterations

    @property
    def W(self):
        return self._W

    @property
    def iterate(self):
        return self._iterate

    @property
    def strides(self):
        return self._strides

    @property
    def padding(self):
        return self._padding

    def get_lipschitz_constant(self, **kwargs):

        def body(i, x):
            x = l2_normalize(x)
            x_p = tf.nn.conv2d(
                x,
                self.W,
                strides=self.strides,
                padding=self.padding)
            x = tf.nn.conv2d_transpose(
                x_p,
                self.W,
                x.shape,
                strides=self.strides,
                padding=self.padding)

            return i + 1, x

        _, x = tf.while_loop(
            self._while_cond, body, [tf.constant(0), self._iterate])

        # Update the power iterate.
        self.iterate.assign(x)

        Wx = tf.nn.conv2d(x, self.W, strides=self.strides, padding=self.padding)

        return tf.sqrt(
            tf.reduce_sum(Wx**2.) / 
            (tf.reduce_sum(x**2.) + gloro.constants.EPS))


class ScalingLayerComputer(LipschitzComputer):
    def __init__(self, layer):
        super().__init__(layer)

        self._w = layer._weight

    @property
    def w(self):
        return self._w

    def get_lipschitz_constant(self, **kwargs):
        return tf.abs(self.w)


class JoinLayerComputer(LipschitzComputer):

    def get_lipschitz_constant(self, lc):
        result = (
            lc[gloro.constants.RESIDUAL_BRANCH] + 
            lc[gloro.constants.SKIP_BRANCH])

        lc[gloro.constants.RESIDUAL_BRANCH] = 1.
        lc[gloro.constants.SKIP_BRANCH] = 1.

        return result


class AveragePoolingComputer(LipschitzComputer):
    def __init__(self, layer):
        super().__init__(layer)

        W = tf.eye(layer.input.shape[-1])[None,None] * (
            tf.ones(layer.pool_size)[:,:,None,None]) / (
                layer.pool_size[0] * layer.pool_size[1])

        x0 = tf.random.truncated_normal(
            shape=(1,*layer.input_shape[1:]))

        def body(i, x):
            x = l2_normalize(x)
            x_p = tf.nn.conv2d(
                x, W,
                strides=layer.strides,
                padding=layer.padding.upper())
            x = tf.nn.conv2d_transpose(
                x_p, W, x.shape,
                strides=layer.strides,
                padding=layer.padding.upper())

            return i + 1, x

        _, x = tf.while_loop(lambda i, _: i < 100, body, [tf.constant(0), x0])

        Wx = tf.nn.conv2d(
            x, W, 
            strides=layer.strides, 
            padding=layer.padding.upper())

        self._lc = tf.sqrt(
            tf.reduce_sum(Wx**2.) / 
            (tf.reduce_sum(x**2.) + gloro.constants.EPS))

    def get_lipschitz_constant(self, **kwargs):
        return self._lc



import numpy as np
import re
import tensorflow as tf

from gloro.utils import get_value


class AffinitySet(object):
    def __init__(self, one_hot_encoding):
        self._mask = tf.constant(one_hot_encoding, dtype='int32')
        self._num_classes = one_hot_encoding.shape[1]

        self._max_set_size = tf.reduce_max(
            tf.reduce_sum(self._mask, axis=1))

    @property
    def num_classes(self):
        return self._num_classes

    @property
    def mask(self):
        return get_value(self._mask)

    @property
    def max_set_size(self):
        return get_value(self._max_set_size)

    def matches(self, top_k_preds):
        asm = self._mask[None,None]
        ccm = self._get_cumulative_class_map(top_k_preds)[:,:,None]

        return tf.reduce_prod(1 - ccm + asm * ccm, axis=3)

    def where_matches(self, top_k_preds):
        asm = self._mask[None,None]
        ccm = self._get_cumulative_class_map(top_k_preds)[:,:,None]

        # ASM : ( -- , -, n, c)
        # CCM : (None, k, -, c)

        # We take the sum (capped at a value of 1) as a logical OR over all 
        # affinity sets because we only care whether a given set of top-k 
        # classes matches with any affinity set.
        where_matches = tf.reduce_sum(
            # We take the product (i.e., a logical AND) over all class positions
            # since the cumulative class mask must be a match for the affinity 
            # set mask for all class positions.
            tf.reduce_prod(
                # This represents the logical formula, CCM_i ==> ASM_i. In other
                # words, the cumulative class mask should only be 1 where the
                # affinity set mask is 1, in order for the corresponding k to be
                # considered a match for that affinity set.
                1 - ccm + asm * ccm, 
                axis=3),
            axis=2) > 0

        return tf.cast(where_matches, 'bool')

    def _get_cumulative_class_map(self, top_k_preds):
        # Assumes `top_k_preds` has shape (None, k, c), where c is the number of
        # classes, and that it represents a one-hot encoding of the kth-highest
        # class for each k.
        return tf.math.cumsum(tf.cast(top_k_preds, 'int32'), axis=1)

    def as_indices(self):
        return [
            [
                int(i) for i, keep in enumerate(affinity_set) if keep
            ]
            for affinity_set in self._mask
        ]

    def __str__(self):
        return str(self.as_indices()).replace(' ', '')[1:-1]

    @staticmethod
    def from_string(s):
        # Expects strings of the form like the following example:
        #
        #   '[0][1,2,3][3,4]'
        #
        return AffinitySet.from_class_indices([
            [int(e) for e in block[1:-1].split(',')]
            for block in re.findall(r'\[.*?\]', s)
        ])

    @staticmethod
    def from_class_indices(class_indices):
        num_classes = max([max(s) for s in class_indices]) + 1

        one_hot_encoding = np.array([
            (np.arange(num_classes)[:,None] == np.array(s)[None]).sum(axis=1)
            for s in class_indices
        ])

        return AffinitySet(one_hot_encoding)

    @staticmethod
    def from_one_hot(one_hot_encoding):
        return AffinitySet(one_hot_encoding)

    def __str__(self):
        sets = '\n'.join([
            f'  {str([i for i, v in enumerate(s) if v == 1])}'
            for s in self._mask.numpy()
        ])
        return f'AffinitySet: [\n{sets}]'

    def __repr__(self):
        return str(self)



import numpy as np
import os
import tensorflow as tf
import tensorflow_datasets as tfds

from tensorflow.data import Dataset
from tensorflow.keras.optimizers import Adam
from tensorflow.keras.optimizers import SGD


def data_augmentation(
        flip=True, 
        saturation=(0.5, 1.2), 
        contrast=(0.8, 1.2), 
        zoom=0.1,
        noise=None):
    
    def augment(x, y):
        batch_size = tf.shape(x)[0]
        input_shape = x.shape[1:]

        # Horizontal flips
        if flip:
            x = tf.image.random_flip_left_right(x)

        # Randomly adjust the saturation and contrast.
        if saturation is not None and input_shape[-1] == 3:
            x = tf.image.random_saturation(
                x, lower=saturation[0], upper=saturation[1])

        if contrast is not None:
            x = tf.image.random_contrast(
                x, lower=contrast[0], upper=contrast[1])

        # Randomly zoom.
        if zoom is not None:
            widths = tf.random.uniform([batch_size], 1. - zoom, 1.)
            top_corners = tf.random.uniform(
                [batch_size, 2], 0, 1. - widths[:, None])
            bottom_corners = top_corners + widths[:, None]
            boxes = tf.concat((top_corners, bottom_corners), axis=1)

            x = tf.image.crop_and_resize(
                x, boxes,
                box_indices=tf.range(batch_size),
                crop_size=input_shape[0:2])

        if noise is not None:
            x = x + tf.random.normal(tf.shape(x), stddev=noise)

        return x, y
    
    return augment


def get_data(dataset, batch_size, augmentation=None):

    # Get the augmentation.
    if augmentation is None or augmentation.lower() == 'none':
        augmentation = data_augmentation(
            flip=False, saturation=None, contrast=None, zoom=None)

    elif augmentation == 'all':
        augmentation = data_augmentation()

    elif augmentation == 'standard':
        augmentation = data_augmentation(
            saturation=None, contrast=None, zoom=0.25)

    elif augmentation == 'no_flip':
        augmentation = data_augmentation(
            flip=False, saturation=None, contrast=None, zoom=0.25)

    elif augmentation.startswith('rs'):
        flip = data_augmentation.startswith('rs_no_flip')

        if augmentation.startswith('rs.'):
            noise = float(augmentation.split('rs.')[1])

        elif augmentation.startswith('rs_no_flip.'):
            noise = float(augmentation.split('rs_no_flip.')[1])

        else:
            noise = 0.125

        augmentation = data_augmentation(
            flip=flip, saturation=None, contrast=None, noise=noise)

    else:
        raise ValueError(f'unknown augmentation type: {augmentation}')

    # Load the data.
    if dataset == 'tiny-imagenet':

        if 'TINY_IMAGENET_LOCATION' in os.environ:
            tiny_imagenet_dir = os.environ['TINY_IMAGENET_LOCATION']

        else:
            raise ValueError(
                'to use tiny-imagenet you must set the '
                '"TINY_IMAGENET_LOCATION" environment variable')

        x_tr = np.load(f'{tiny_imagenet_dir}/TinyImagenet_train_x.npy')
        x_te = np.load(f'{tiny_imagenet_dir}/TinyImagenet_test_x.npy')
        y_tr = np.load(f'{tiny_imagenet_dir}/TinyImagenet_train_y.npy')
        y_te = np.load(f'{tiny_imagenet_dir}/TinyImagenet_test_y.npy')

        train = (Dataset.from_tensor_slices((x_tr, y_tr))
            .cache()
            .batch(batch_size)
            .map(
                augmentation, 
                num_parallel_calls=tf.data.experimental.AUTOTUNE,
                deterministic=False)
            .prefetch(tf.data.experimental.AUTOTUNE))

        test = (Dataset.from_tensor_slices((x_te, y_te))
            .cache()
            .batch(batch_size)
            .prefetch(tf.data.experimental.AUTOTUNE))

        metadata = lambda: None
        metadata.features = {
            'image': lambda: None,
            'label': lambda: None,
        }
        metadata.features['label'].num_classes = 200
        metadata.features['image'].shape = (64, 64, 3)

    else: 
        tfds_dir = os.environ['TFDS_DIR'] if 'TFDS_DIR' in os.environ else None

        split = ['train', 'test']

        (train, test), metadata = tfds.load(
            dataset,
            data_dir=tfds_dir,
            split=split, 
            with_info=True, 
            shuffle_files=True, 
            as_supervised=True)

        train = (train
            .map(
                lambda x,y: (tf.cast(x, 'float32') / 255., y), 
                num_parallel_calls=tf.data.experimental.AUTOTUNE,
                deterministic=False)
            .cache()
            .batch(batch_size)
            .map(
                augmentation, 
                num_parallel_calls=tf.data.experimental.AUTOTUNE,
                deterministic=False)
            .prefetch(tf.data.experimental.AUTOTUNE))

        test = (test
            .map(
                lambda x,y: (tf.cast(x, 'float32') / 255., y), 
                num_parallel_calls=tf.data.experimental.AUTOTUNE,
                deterministic=False)
            .cache()
            .batch(batch_size)
            .prefetch(tf.data.experimental.AUTOTUNE))

    return train, test, metadata


def get_optimizer(optimizer, lr):
    if optimizer == 'adam':
        return Adam(lr=lr)

    elif optimizer.startswith('sgd'):
        if optimizer.startswith('sgd.'):
            return SGD(lr=lr, momentum=float(optimizer.split('sgd.')[1]))

        else:
            return SGD(lr=lr)

    else:
        raise ValueError(f'unknown optimizer: {optimizer}')



import tensorflow as tf


def rtk_vra_sparse(y_true, y_pred):
    labels = tf.cast(y_true, 'int32')[:,0]

    bot_index = tf.shape(y_pred)[1] - 2

    guarantee = tf.cast(y_pred[:, -1], 'int32')

    k = tf.reduce_max(guarantee)

    y_j, _ = tf.math.top_k(y_pred[:,:-2], k=k)

    pred_at_ground_truth = tf.gather(y_pred, labels, batch_dims=1)

    minimum_pred = tf.gather(y_j, guarantee - 1, batch_dims=1)

    return tf.reduce_mean(
        tf.cast(pred_at_ground_truth >= minimum_pred, 'float32') *
        tf.cast(
            tf.not_equal(
                tf.argmax(y_pred[:,:-1], axis=1, output_type='int32'),
                bot_index),
            'float32'))

def rtk_vra_cat(y_true, y_pred):
    labels = tf.argmax(y_true, axis=1, output_type='int32')[:,None]

    return rtk_vra_sparse(labels, y_pred)


def rtk_vra(y_true, y_pred):
    return tf.case([
        (tf.equal(tf.shape(y_true)[1], 1),
            lambda: rtk_vra_sparse(y_true, y_pred))],
        default=lambda: rtk_vra_cat(y_true, y_pred))


def affinity_vra_sparse(y_true, y_pred): return rtk_vra_sparse(y_true, y_pred)
def affinity_vra_cat(y_true, y_pred): return rtk_vra_cat(y_true, y_pred)
def affinity_vra(y_true, y_pred): return rtk_vra(y_true, y_pred)


def top_k_clean_acc(k, sparse=None):

    def metric_sparse(y_true, y_pred):
        return sparse_top_k_categorical_accuracy(y_true, y_pred[:, :-1], k)

    metric_sparse.__name__ = f'top_{k}_clean_acc_sparse'

    def metric_cat(y_true, y_pred):
        return top_k_categorical_accuracy(y_true, y_pred[:, :-1], k)

    metric_cat.__name__ = f'top_{k}_clean_acc_cat'

    def metric(y_true, y_pred):
        return tf.case([
            (tf.equal(tf.shape(y_true)[1], 1),
                lambda: metric_sparse(y_true, y_pred))],
            default=lambda: metric_cat(y_true, y_pred))

    metric.__name__ = f'top_{k}_clean_acc'

    if sparse is None:
        return metric

    else:
        return metric_sparse if sparse else metric_cat

def top_k_clean_acc_sparse(k):
    return top_k_clean_acc(k, sparse=True)

def top_k_clean_acc_cat(k):
    return top_k_clean_acc(k, sparse=False)


import numpy as np
import re


def _linear_interp(start, end, duration):
    return [
        start + (end - start) * i / duration for i in range(duration)
    ]

def _exp_interp(start, end, duration):
    B = np.log(end / start) / duration

    return [start * np.exp(B * i) for i in range(duration)]

def _log_interp(start, end, duration):
    return [
        np.log((np.exp(end) - np.exp(start)) / duration * i + np.exp(start))
        for i in range(duration)
    ]

def _no_interp(start, duration):
    return [start for _ in range(duration)]


def _parse_section(
    section, duration, is_start=False, is_end=False, interpolated=False
):
    if not (section.startswith('[') and section.endswith(']')):
        raise ValueError(f'Invalid section: {section}')

    section = section[1:-1]

    if not (is_start or is_end or ':' in section):
        raise ValueError(f'Invalid section contents: {section}')

    if is_start:
        time = 0

    elif is_end:
        time = duration

    if ':' in section:
        time, value = section.split(':')

        if time.endswith('%'):
            time = int(float(time[:-1]) / 100. * duration)
        else:
            time = int(time)

    else:
        value = section

    if interpolated:
        value = float(value)

    else:
        try:
            value = int(value)
        except:
            try:
                value = float(value)
            except:
                pass

    return time, value


class Schedule(object):

    def __init__(self, schedule_string, duration, base_value):
        if schedule_string == 'fixed':
            self._schedule = [base_value]

        elif schedule_string == 'linear':
            self._schedule = Schedule(
                '[0.]-[1.]', duration, base_value).as_list()

        elif schedule_string == 'linear_half':
            self._schedule = Schedule(
                '[0.]-[50%:1.]-[1.]', duration, base_value).as_list()

        elif schedule_string.startswith('linear_from_'):
            start = float(schedule_string.split('linear_from_')[1]) / base_value
            self._schedule = Schedule(
                f'[{start}]-[1.]', duration, base_value).as_list()

        elif schedule_string.startswith('linear_half_from_'):
            start = float(
                schedule_string.split('linear_half_from_')[1]) / base_value
            self._schedule = Schedule(
                f'[{start}]-[50%:1.]-[1.]', duration, base_value).as_list()

        elif schedule_string == 'logarithmic':
            self._schedule = Schedule(
                '[0.01]-log-[1.]', duration, base_value).as_list()

        elif schedule_string == 'logarithmic_half':
            self._schedule = Schedule(
                '[0.01]-log-[50%:1.]-[1.]', duration, base_value).as_list()

        elif schedule_string.startswith('decay_to_'):
            end = float(schedule_string.split('decay_to_')[1]) / base_value
            self._schedule = Schedule(
                f'[1.]-exp-[{end}]', duration, base_value).as_list()

        elif schedule_string.startswith('decay_after_half_to_'):
            end = float(
                schedule_string.split('decay_after_half_to_')[1]) / base_value
            self._schedule = Schedule(
                f'[1.]-[50%:1.]-exp-[{end}]', duration, base_value).as_list()

        elif schedule_string.startswith('decay_until_half_to_'):
            end = float(
                schedule_string.split('decay_until_half_to_')[1]) / base_value
            self._schedule = Schedule(
                f'[1.]-exp-[50%:{end}]-[{end}]', duration, base_value).as_list()

        else:
            transitions_and_sections = [
                (transition, section)
                for transition, section in
                zip(
                    re.split(r'\[.*?\]', schedule_string),
                    re.findall(r'\[.*?\]', schedule_string))
            ]

            n = len(transitions_and_sections)

            uses_interpolation = not np.all(
                np.array([
                    transition for transition, _ in transitions_and_sections
                ]) == '')

            _time = 0
            self._schedule = []
            for i in range(n - 1):
                start_section = transitions_and_sections[i][1]
                transition, end_section = transitions_and_sections[i+1]

                start_time, start_val = _parse_section(
                    start_section,
                    duration,
                    is_start=i == 0,
                    is_end=False,
                    interpolated=uses_interpolation)
                end_time, end_val = _parse_section(
                    end_section,
                    duration,
                    is_start=False,
                    is_end=i + 1 == n - 1,
                    interpolated=uses_interpolation)

                if not (_time <= start_time <= end_time):
                    raise ValueError('Invalid time specification')

                start_val *= base_value
                end_val *= base_value

                if transition == '-':
                    self._schedule += _linear_interp(
                        start_val, end_val, end_time - start_time)

                elif transition == '-exp-':
                    self._schedule += _exp_interp(
                        start_val, end_val, end_time - start_time)

                elif transition == '-log-':
                    self._schedule += _log_interp(
                        start_val, end_val, end_time - start_time)

                else:
                    self._schedule += _no_interp(
                        start_val, end_time - start_time)

                    if i + 1 == n - 1 and end_time < duration - 1:
                        self._schedule.append(end_val)
            
    def __getitem__(self, index):
        return self._schedule[min(index, len(self._schedule) - 1)]

    def __len__(self):
        return len(self._schedule)

    def as_list(self):
        return self._schedule



from gloro.layers.network_layers import MinMax
from gloro.layers.network_layers import InvertibleDownsampling
from gloro.layers.network_layers import Scaling
from gloro.layers.network_layers import Bias
from gloro.layers.network_layers import ResnetBlock


import numpy as np
import tensorflow as tf

from gloro.layers.margin_layers import LipschitzMargin
from gloro.utils import get_value
from gloro.utils import set_value


class LipschitzMarginRtk(LipschitzMargin):
    def __init__(
        self, 
        epsilon,
        k,
        model=None,
        num_iterations=5, 
        norm='l2', 
        strictest_guarantee=False,
        **kwargs,
    ):
        super().__init__(
            epsilon, 
            model=model,
            num_iterations=num_iterations, 
            norm=norm, 
            **kwargs)

        self._k = tf.Variable(k, name='k', trainable=False)
        self._strictest_guarantee = tf.Variable(
            strictest_guarantee, name='strict', trainable=False)

    @property
    def k(self):
        return get_value(self._k)

    @k.setter
    def k(self, new_k):
        set_value(self._k, new_k)

    @property
    def strictest_guarantee(self):
        return get_value(self._strictest_guarantee)

    @strictest_guarantee.setter
    def strictest_guarantee(self, new_strictest_guarantee):
        set_value(self._strictest_guarantee, new_strictest_guarantee)

    def _get_yj_j_Kij(self, y):

        kW = self._kW()

        y_j, j = tf.math.top_k(y, k=self._k)
        where_not_j = tf.not_equal(y[:,None], y_j[:,:,None])

        # Get the weight column of the predicted class.
        kW_j = tf.gather(tf.transpose(kW), j)

        # Get weights that predict the value y_j - y_i for all i != j.
        kW_ij = kW_j[:,:,:,None] - kW[None,None]

        # We do this instead of `tf.linalg.norm(W_d)` because of an apparent bug
        # in `tf.linalg.norm` that leads to NaN values.
        K_ij = tf.sqrt(tf.reduce_sum(kW_ij * kW_ij, axis=2))

        return y_j, j, K_ij

    def _lipschitz_constant_fn(self, y):

        y_j, j, K_ij = self._get_yj_j_Kij(y)

        return K_ij + tf.math.cumsum(
            tf.where(
                tf.equal(y[:,None], y_j[:,:,None]), 
                -np.inf, 
                0.), 
            axis=1, 
            reverse=True)
      
    def _get_mk(self, y):
        y_j, j, K_ij = self._get_yj_j_Kij(y)
        
        where_in_top_k = tf.equal(y[:,None], y_j[:,:,None])
        where_yi_lt_yjk = tf.math.cumsum(
            tf.where(where_in_top_k, -np.inf, 0.), 
            axis=1)
        where_l_leq_k = tf.where(
            tf.equal(tf.cumsum(tf.eye(tf.cast(self._k, 'uint8'))), 0.),
            -np.inf,
            0)
        
        mk_j = y_j[:,None] - tf.reduce_max(
          y[:,None,None] + self._epsilon * K_ij[:,None] + 
              where_yi_lt_yjk[:,:,None] + 
              where_l_leq_k[None,:,:,None],
          axis=3)
        
        return tf.reduce_min(mk_j, axis=2)
      
    def call(self, y):
      
        mk = self._get_mk(y)
        
        m = tf.reduce_max(mk, axis=1)

        all_guarantees = tf.cast(mk > 0, 'int32')

        # For the purpose of metrics, get the most lax (i.e., largest) k that we
        # are top-k robust with.
        loosest_k = tf.cast(
            1 + tf.argmax(
                tf.cast(mk > 0, 'float32') +
                    # Adding this breaks ties in favor of larger k.
                    tf.cast(tf.range(self._k), 'float32') * 1e-7,
                axis=1),
            'float32')[:,None]

        y_bot = tf.reduce_max(y, axis=1) - m

        return tf.concat([y, y_bot[:,None]], axis=1), all_guarantees, loosest_k


class LipschitzMarginAffinity(LipschitzMarginRtk):
    def __init__(
        self, 
        epsilon,
        k,
        affinity_sets,
        model=None,
        num_iterations=5, 
        norm='l2', 
        **kwargs,
    ):
        super().__init__(
            epsilon, 
            k=k,
            model=model,
            num_iterations=num_iterations, 
            norm=norm, 
            **kwargs)

        self._affinity_sets = affinity_sets

    @property
    def k(self):
        return super().k

    @k.setter
    def k(self, new_k):
        raise RuntimeError('`k` is not settable')

    @property
    def affinity_sets(self):
        return self._affinity_sets

    def _get_mk(self, y):
        y_j, j, K_ij = self._get_yj_j_Kij(y)
        
        where_in_top_k = tf.equal(y[:,None], y_j[:,:,None])
        where_yi_lt_yjk = tf.math.cumsum(
            tf.where(where_in_top_k, -np.inf, 0.), 
            axis=1)
        where_l_leq_k = tf.where(
            tf.equal(tf.cumsum(tf.eye(tf.cast(self._k, 'uint8'))), 0.),
            -np.inf,
            0)
        
        mk_j = y_j[:,None] - tf.reduce_max(
          y[:,None,None] + self._epsilon * K_ij[:,None] + 
              where_yi_lt_yjk[:,:,None] + 
              where_l_leq_k[None,:,:,None],
          axis=3)
        
        mk = tf.reduce_min(mk_j, axis=2)

        # We don't want to consider any margins that correspond to a k that
        # corresponds to a set of classes that do not match any of our affinity
        # sets.
        return tf.where(
            self._affinity_sets.where_matches(where_in_top_k),
            mk,
            -np.inf)



import json
import os
import shutil
import tarfile
import tensorflow as tf

from tensorflow.keras.models import Model

import gloro

from gloro.layers.margin_layers import LipschitzMargin
from gloro.training.callbacks import UpdatePowerIterates
from gloro.training.losses import get as get_loss


class GloroNet(Model):
    def __init__(
        self, 
        inputs=None, 
        outputs=None, 
        epsilon=None, 
        num_iterations=5,
        model=None, 
        _skip_init=False,
        **kwargs,
    ):
        if _skip_init:
            # We have to do this because keras requires __init__ to be called on
            # *all* subclasses.
            super().__init__(inputs, outputs, **kwargs)
            return

        if epsilon is None:
            raise ValueError('`epsilon` is required')
        
        if model is None and (inputs is None or outputs is None):
            raise ValueError(
                'must specify either `inputs` and `outputs` or `model`')
            
        if model is not None and inputs is not None and outputs is not None:
            raise ValueError(
                'cannot specify both `inputs` and `outputs` and `model`')

        if inputs is None or outputs is None:
            inputs = model.inputs
            outputs = model.outputs
            
        else:
            model = Model(inputs, outputs)

        if not isinstance(outputs, (list, tuple)):
            outputs = [outputs]
            
        if len(outputs) > 1:
            raise ValueError(
                f'only models with a single output are supported, but got '
                f'{len(outputs)} outputs')

        margin_layer = LipschitzMargin(
            epsilon, model, num_iterations=num_iterations)

        margin = margin_layer(*outputs)
        
        super().__init__(inputs, margin, **kwargs)

        self._inputs = inputs
        self._output = outputs[0]
        self._margin_layer = margin_layer
        self._f = GloroNet._ModelContainer(model)

    @property
    def epsilon(self):
        return self._margin_layer.epsilon

    @epsilon.setter
    def epsilon(self, new_epsilon):
        self._margin_layer.epsilon = new_epsilon

    @property
    def num_iterations(self):
        return self._margin_layer.num_iterations

    @num_iterations.setter
    def num_iterations(self, new_num_iterations):
        self._margin_layer.num_iterations = new_num_iterations

    @property
    def f(self):
        return self._f.model


    def lipschitz_constant(self, X=None, y=None):
        if X is None and y is None:
            return self._margin_layer.lipschitz_constant(
                tf.eye(self._output.shape[1]))

        elif X is not None:
            if not isinstance(X, (list, tuple)):
                X = [X]

            return self._margin_layer.lipschitz_constant(self.f(X))

        else:
            return self._margin_layer.lipschitz_constant(y)

    def predict_clean(self, *args, **kwargs):
        return self._f.model.predict(*args, **kwargs)

    def freeze_lipschitz_constant(
        self, 
        converge=True, 
        convergence_threshold=1e-4, 
        max_tries=1000, 
        batch_size=None,
        verbose=False,
    ):
        if converge:
            self.refresh_iterates(
                converge=converge,
                convergence_threshold=convergence_threshold,
                max_tries=max_tries,
                batch_size=batch_size,
                verbose=verbose)

        self._margin_layer.freeze()
        self.trainable = False

        return self

    def refresh_iterates(
        self, 
        converge=True, 
        convergence_threshold=1e-4, 
        max_tries=1000, 
        batch_size=None,
        verbose=False,
    ):
        self._margin_layer.refresh_iterates(
            converge=converge,
            convergence_threshold=convergence_threshold,
            max_tries=max_tries,
            batch_size=batch_size,
            verbose=verbose)

        return self

    def update_iterates(self):
        self._margin_layer.update_iterates()

        return self


    # Overriding `keras.Model` functions.

    def compile(self, **kwargs):
        if 'loss' in kwargs:
            if isinstance(kwargs['loss'], str):
                kwargs['loss'] = get_loss(kwargs['loss'])

        if 'metrics' in kwargs:
            metrics = kwargs['metrics']
            new_metrics = []

            for metric in metrics:
                if metric in gloro.constants.GLORO_CUSTOM_OBJECTS:
                    metric = gloro.constants.GLORO_CUSTOM_OBJECTS[metric]

                new_metrics.append(metric)

            kwargs['metrics'] = new_metrics

        super().compile(**kwargs)

    def fit(self, *args, update_iterates=True, **kwargs):
        if update_iterates:
            includes_update_iterates = False

            if 'callbacks' in kwargs:
                callbacks = kwargs['callbacks']

                for callback in callbacks:
                    if isinstance(callback, UpdatePowerIterates):
                        includes_update_iterates = True

            else:
                callbacks = []

            if not includes_update_iterates:
                kwargs['callbacks'] = [
                    UpdatePowerIterates(verbose=True),
                    *callbacks
                ]
        
        return super().fit(*args, **kwargs)

    def save(self, file_name, overwrite=True):
        if file_name.endswith('.h5'):
            file_name = file_name[:-3]

        elif file_name.endswith('.gloronet'):
            file_name = file_name[:-9]

        try:
            os.mkdir(file_name)

            self.f.save(f'{file_name}/f.h5')

            with open(f'{file_name}/config.json', 'w') as json_file:
                json.dump(self.get_config(), json_file)

            with tarfile.open(f'{file_name}.gloronet', 'w:gz') as tar:
                tar.add(file_name, arcname=os.path.basename(file_name))

        finally:
            shutil.rmtree(file_name)

    @classmethod
    def load_model(cls, file_name, custom_objects={}, converge=True):
        custom_objects = dict(gloro.constants.GLORO_CUSTOM_OBJECTS.copy(), **custom_objects)

        if file_name.endswith('.h5'):
            file_name = file_name[:-3]

        elif file_name.endswith('.gloronet'):
            file_name = file_name[:-9]
            
        temp_dir = f'{file_name}___'
        
        try:
            os.mkdir(temp_dir)
            
            with tarfile.open(f'{file_name}.gloronet', 'r:gz') as tar:

                for member in tar:
                    if member.name.endswith('.h5'):
                        tar.extract(member, f'{temp_dir}')

                        model = tf.keras.models.load_model(
                            f'{temp_dir}/{member.name}',
                            custom_objects=custom_objects)

                    elif member.name.endswith('.json'):
                        with tar.extractfile(member) as f:
                            config = json.load(f)
        finally:
            shutil.rmtree(temp_dir)

        if converge:
            return cls(model=model, **config).refresh_iterates()
        else:
            return cls(model=model, **config)

    def get_config(self):
        return {
            'epsilon': float(self.epsilon),
            'num_iterations': int(self.num_iterations),
        }


    class _ModelContainer(object):
        def __init__(self, model):
            self.model = model



import numpy as np
import tensorflow as tf

from tensorflow.keras.losses import BinaryCrossentropy
from tensorflow.keras.losses import CategoricalCrossentropy
from tensorflow.keras.losses import KLDivergence
from tensorflow.keras.losses import Loss
from tensorflow.keras.losses import SparseCategoricalCrossentropy

from gloro.utils import add_extra_column
from gloro.utils import get_value
from gloro.utils import set_value


class Crossentropy(Loss):

    def __init__(self, sparse=False, temperature=1., **kwargs):
        super().__init__(**kwargs)

        kwargs['from_logits'] = True

        if sparse:
            self._loss = SparseCategoricalCrossentropy(**kwargs)
        else:
            self._loss = CategoricalCrossentropy(**kwargs)

        self._sparse = sparse

        self._temperature = tf.Variable(temperature, trainable=False)

    @property
    def temperature(self):
        return get_value(self._temperature)

    @temperature.setter
    def temperature(self, new_temperature):
        set_value(self._temperature, new_temperature)

    def call(self, y_true, y_pred):
        # Add an extra column of zeros for the bottom class.
        if not self._sparse:
            y_true = add_extra_column(y_true)

        return self._loss.call(y_true, y_pred / self._temperature)

    def get_config(self):
        config = self._loss.get_config()
        config.update({
            'sparse': self._sparse,
            'temperature': self.temperature,
        })
        return config


class Trades(Loss):

    def __init__(
        self, 
        lam=1., 
        sparse=False, 
        robust_loss_string='kl_divergence', 
        temperature=1.,
        **kwargs,
    ):
        super().__init__(**kwargs)

        self._lam = tf.Variable(lam, trainable=False)

        self._robust_loss_string = robust_loss_string

        if 'from_logits' in kwargs:
            del kwargs['from_logits']

        if sparse:
            self._crossentropy = SparseCategoricalCrossentropy(
                from_logits=True, **kwargs)
        else:
            self._crossentropy = CategoricalCrossentropy(
                from_logits=True, **kwargs)

        if robust_loss_string is None or robust_loss_string == 'crossentropy':
            self._robust_loss = CategoricalCrossentropy(
                from_logits=False, **kwargs)

        elif robust_loss_string == 'kl_divergence':
            self._robust_loss = KLDivergence(**kwargs)

        else:
            raise ValueError(f'unsupported robust loss: {robust_loss_string}')

        self._sparse = sparse

        self._temperature = tf.Variable(temperature, trainable=False)

    @property
    def lam(self):
        return get_value(self._lam)

    @lam.setter
    def lam(self, new_lam):
        set_value(self._lam, new_lam)

    @property
    def temperature(self):
        return get_value(self._temperature)

    @temperature.setter
    def temperature(self, new_temperature):
        set_value(self._temperature, new_temperature)

    def call(self, y_true, y_pred):
        # Add an extra column of zeros for the bottom class.
        if not self._sparse:
            y_true = add_extra_column(y_true)

        # Encourage predicting the correct class, even non-robustly.
        if self._sparse:
            standard_loss = self._crossentropy(y_true, y_pred[:, :-1])
        else:
            standard_loss = self._crossentropy(y_true[:, :-1], y_pred[:, :-1])

        # Encourage predicting robustly, even incorrectly. We take the robust
        # loss but using the model's prediction as the ground truth.
        y_pred_soft = tf.nn.softmax(y_pred)

        new_ground_truth = add_extra_column(tf.nn.softmax(y_pred[:, :-1]))

        robust_loss = self._robust_loss(
            new_ground_truth, y_pred_soft / self._temperature)

        # Combine the standard and robust terms.
        return standard_loss + self._lam * robust_loss

    def get_config(self):
        return {
            'lam': self.lam,
            'sparse': self._sparse,
            'robust_loss_string': self._robust_loss_string,
            'temperature': self.temperature,
        }


def get(loss):
    sparse = False
    if loss.startswith('sparse_'):
        loss = loss.split('sparse_')[1]
        sparse = True

    if loss == 'crossentropy':
        return Crossentropy(sparse=sparse)

    elif loss.startswith('trades_ce'):
        if loss.startswith('trades_ce.'):
            return Trades(
                float(loss.split('trades_ce.')[1]), 
                robust_loss_string='crossentropy',
                sparse=sparse)

        else:
            return Trades(robust_loss_string='crossentropy', sparse=sparse)

    elif loss.startswith('trades_kl'):
        if loss.startswith('trades_kl.'):
            return Trades(
                float(loss.split('trades_kl.')[1]), 
                robust_loss_string='kl_divergence',
                sparse=sparse)
        else:
            return Trades(robust_loss_string='kl_divergence', sparse=sparse)

    else:
        raise ValueError(f'unknown loss function: {loss}')



from gloro.layers import Bias
from gloro.layers import InvertibleDownsampling
from gloro.layers import MinMax
from gloro.layers import Scaling
from gloro.relaxations.metrics import affinity_vra
from gloro.relaxations.metrics import affinity_vra_cat
from gloro.relaxations.metrics import affinity_vra_sparse
from gloro.relaxations.metrics import rtk_vra
from gloro.relaxations.metrics import rtk_vra_cat
from gloro.relaxations.metrics import rtk_vra_sparse
from gloro.relaxations.metrics import top_k_clean_acc
from gloro.relaxations.metrics import top_k_clean_acc_cat
from gloro.relaxations.metrics import top_k_clean_acc_sparse
from gloro.training.losses import Crossentropy
from gloro.training.losses import Trades
from gloro.training.metrics import clean_acc
from gloro.training.metrics import clean_acc_cat
from gloro.training.metrics import clean_acc_sparse
from gloro.training.metrics import vra
from gloro.training.metrics import vra_cat
from gloro.training.metrics import vra_sparse


GLORO_CUSTOM_OBJECTS = {
    'Crossentropy': Crossentropy,
    'Trades': Trades,
    'clean_acc': clean_acc,
    'clean_acc_cat': clean_acc_cat,
    'clean_acc_sparse': clean_acc_sparse,
    'vra': vra,
    'vra_cat': vra_cat,
    'vra_sparse':vra_sparse,

    'top_k_clean_acc': top_k_clean_acc,
    'top_k_clean_acc_cat': top_k_clean_acc_cat,
    'top_k_clean_acc_sparse': top_k_clean_acc_sparse,
    'rtk_vra': rtk_vra,
    'rtk_vra_cat': rtk_vra_cat,
    'rtk_vra_sparse': rtk_vra_sparse,
    'affinity_vra': affinity_vra,
    'affinity_vra_cat': affinity_vra_cat,
    'affinity_vra_sparse': affinity_vra_sparse,

    'MinMax': MinMax,
    'InvertibleDownsampling': InvertibleDownsampling,
    'Scaling': Scaling,
    'Bias': Bias,
}

EPS = 1e-9

MAIN_BRANCH = 'main'
RESIDUAL_BRANCH = 'residual'
SKIP_BRANCH = 'skip'



if __name__ == '__main__':
    import os

    # Stop tensorflow's obnoxious logging.
    os.environ['TF_CPP_MIN_LOG_LEVEL'] = '3'

import json
import tensorflow as tf
import tensorflow_datasets as tfds

from scriptify import scriptify
from tensorflow.keras.losses import SparseCategoricalCrossentropy
from tensorflow.keras.models import Model
from tensorflow.keras.optimizers import Adam, SGD
from tensorflow.keras.optimizers import SGD

from gloro import GloroNet
from gloro.training import losses
from gloro.training.callbacks import EpsilonScheduler 
from gloro.training.callbacks import LrScheduler
from gloro.training.callbacks import TradesScheduler
from gloro.training.metrics import clean_acc
from gloro.training.metrics import vra
from gloro.training.metrics import rejection_rate
from gloro.utils import print_if_verbose

import architectures
from utils import get_data
from utils import get_optimizer


def train_gloro(
    dataset,
    architecture,
    epsilon,
    power_iterations=5,
    epsilon_schedule='fixed',
    loss='crossentropy',
    augmentation='standard',
    epochs=None,
    batch_size=None,
    optimizer='adam',
    lr=1e-3,
    lr_schedule='fixed',
    trades_schedule=None,
    verbose=True,
):
    _print = print_if_verbose(verbose)

    # Load data and set up data pipeline.
    _print('loading data...')

    train, test, metadata = get_data(dataset, batch_size, augmentation)

    input_shape = metadata.features['image'].shape
    num_classes = metadata.features['label'].num_classes

    # Create the model.
    _print('creating model...')

    try:
        _orig_architecture = architecture
        params = '{}'

        if '.' in architecture:
            architecture, params = architecture.split('.', 1)

        architecture = getattr(architectures, architecture)(
            input_shape, num_classes, **json.loads(params))

    except:
        raise ValueError(f'unknown architecture: {_orig_architecture}')

    g = GloroNet(*architecture, epsilon, num_iterations=power_iterations)

    if verbose:
        g.summary()

    # Compile and train the model.
    _print('compiling model...')

    g.compile(
        loss=losses.get(loss),
        optimizer=get_optimizer(optimizer, lr), 
        metrics=[clean_acc, vra, rejection_rate])

    g.fit(
        train,
        epochs=epochs, 
        validation_data=test,
        callbacks=[
            EpsilonScheduler(epsilon_schedule),
            LrScheduler(lr_schedule),
        ] + ([TradesScheduler(trades_schedule)] if trades_schedule else []))

    return g



if __name__ == '__main__':

    @scriptify
    def script(
        dataset,
        architecture,
        epsilon,
        power_iterations=5,
        epsilon_schedule='fixed',
        loss='crossentropy',
        augmentation='standard',
        epochs=100,
        batch_size=128,
        optimizer='adam',
        lr=1e-3,
        lr_schedule='decay_to_0.000001',
        trades_schedule=None,
        gpu=0,
    ):
        # Select the GPU and allow memory growth to avoid taking all the RAM.
        gpus = tf.config.experimental.list_physical_devices('GPU')
        tf.config.experimental.set_visible_devices(gpus[gpu], 'GPU')
        device = gpus[gpu]

        for device in tf.config.experimental.get_visible_devices('GPU'):
            tf.config.experimental.set_memory_growth(device, True)

        g = train_gloro(
            dataset,
            architecture,
            epsilon,
            power_iterations=power_iterations,
            epsilon_schedule=epsilon_schedule,
            loss=loss,
            augmentation=augmentation,
            epochs=epochs,
            batch_size=batch_size,
            optimizer=optimizer,
            lr=lr,
            lr_schedule=lr_schedule,
            trades_schedule=trades_schedule)

        # Evaluate the model.
        train, test, metadata = get_data(dataset, batch_size, augmentation)

        train_eval = g.evaluate(train)
        test_eval = g.evaluate(test)

        results = {}

        results.update({
            'test_' + metric.name.split('pred_')[-1]: round(value, 4)
            for metric, value in zip(g.metrics, test_eval)
        })
        results.update({
            'train_' + metric.name.split('pred_')[-1]: round(value, 4)
            for metric, value in zip(g.metrics, train_eval)
        })

        print(results)

        return results



import numpy as np

from abc import abstractmethod
from tensorflow.keras.callbacks import Callback
from tensorflow.keras.callbacks import LearningRateScheduler
from time import time

from gloro.training.losses import Trades
from gloro.training.schedules import Schedule
from gloro.utils import get_value
from gloro.utils import print_if_verbose
from gloro.utils import set_value


class VariableScheduler(Callback):

    def __init__(
        self, 
        variable_name,
        schedule_string='fixed', 
        test_value=None,
        base_value=None, 
        interpolate=True,
        verbose=True,
    ):
        super().__init__()

        self._name = variable_name

        self._schedule_string = schedule_string
        self._interpolate_between_epochs = interpolate

        self._base_value = base_value
        self._test_value = test_value

        self._print = print_if_verbose(verbose)

    @abstractmethod
    def get_var(self):
        raise NotImplementedError

    @abstractmethod
    def set_var(self, value):
        raise NotImplementedError

    def on_train_begin(self, logs=None):

        self._num_steps = self.params['steps']

        self._schedule = Schedule(
            self._schedule_string, 
            duration=self.params['epochs'], 
            base_value=(
                self._base_value if self._base_value is not None else
                self.get_var()))

        if self._test_value is None:
            self._test_value = self.get_var()

        self.__prev = None
        self.__current = self._schedule[0]

    def on_test_begin(self, logs=None):
        self._train_value = self.get_var()
        self.set_var(self._test_value)

    def on_test_end(self, logs=None):
        self.set_var(self._train_value)

    def on_epoch_begin(self, epoch, logs=None):
        if epoch < len(self._schedule):
            next_value = self._schedule[epoch]

            self.set_var(next_value)
        
            if self.__prev is None or self.__prev != self.__current:
                self._print(f'---- setting {self._name}={next_value:.5f} ----')

            self.__prev = self.__current
            self.__current = next_value

    def on_train_batch_begin(self, batch, logs=None):
        if self._interpolate_between_epochs:
            self.set_var(
                self.__prev + (self.__current - self.__prev) * (
                    batch / self._num_steps))

    def on_train_end(self, logs=None):
        self.set_var(self._test_value)


class EpsilonScheduler(VariableScheduler):

    def __init__(
        self,
        schedule_string='fixed', 
        test_value=None,
        base_value=None, 
        interpolate=True,
        verbose=True,
    ):
        super().__init__(
            'epsilon', 
            schedule_string=schedule_string,
            test_value=test_value,
            base_value=base_value,
            interpolate=interpolate,
            verbose=verbose)

    def get_var(self):
        return self.model.epsilon

    def set_var(self, value):
        self.model.epsilon = value


class TradesScheduler(VariableScheduler):

    def __init__(
        self,
        schedule_string='fixed', 
        test_value=None,
        base_value=None, 
        interpolate=True,
        verbose=True,
    ):
        super().__init__(
            'TRADES parameter', 
            schedule_string=schedule_string,
            test_value=test_value,
            base_value=base_value,
            interpolate=interpolate,
            verbose=verbose)

    def get_var(self):
        if isinstance(self.model.loss, dict):
            loss = self.model.loss['pred']
        else:
            loss = self.model.loss

        # Make sure that the model is actually using TRADES loss.
        if not isinstance(loss, Trades):
            raise ValueError(
                '`TradesScheduler` can only be used with `Trades` loss.')

        return loss.lam

    def set_var(self, value):
        if isinstance(self.model.loss, dict):
            loss = self.model.loss['pred']
        else:
            loss = self.model.loss

        # Make sure that the model is actually using TRADES loss.
        if not isinstance(loss, Trades):
            raise ValueError(
                '`TradesScheduler` can only be used with `Trades` loss.')

        loss.lam = value


class LrScheduler(VariableScheduler):

    def __init__(
        self,
        schedule_string='fixed', 
        test_value=None,
        base_value=None, 
        interpolate=True,
        verbose=True,
    ):
        super().__init__(
            'learning rate', 
            schedule_string=schedule_string,
            test_value=test_value,
            base_value=base_value,
            interpolate=interpolate,
            verbose=verbose)

    def get_var(self):
        return get_value(self.model.optimizer.learning_rate)

    def set_var(self, value):
        set_value(self.model.optimizer.learning_rate, value)


class UpdatePowerIterates(Callback):
    def __init__(
        self, 
        convergence_threshold=1e-4,
        short_convergence_threshold=1e-2,
        iteration_batch_size=100, 
        do_initial_convergence=True,
        verbose=True,
    ):
        super().__init__()

        self._convergence_threshold = convergence_threshold
        self._short_convergence_threshold = short_convergence_threshold
        self._batch_size = iteration_batch_size
        self._do_initial_convergence = do_initial_convergence
        self._verbose = verbose

        self._print = print_if_verbose(verbose)

        self._converged = False

    def on_train_begin(self, logs=None):

        if self._do_initial_convergence:

            self._print('---- refreshing iterates ----')

            start = time()

            self.model.refresh_iterates(
                convergence_threshold=self._short_convergence_threshold,
                batch_size=self._batch_size,
                verbose=self._verbose)

            self._print(f'   > done: {(time() - start):.2f} seconds')

    def on_test_begin(self, logs=None):
        self._print('\n---- refreshing iterates ----')

        start = time()

        self.model.refresh_iterates(
            convergence_threshold=self._short_convergence_threshold,
            batch_size=self._batch_size,
            verbose=self._verbose)

        self._print(f'   > done: {(time() - start):.2f} seconds')

        self._converged = True

    def on_epoch_begin(self, epoch, logs=None):
        self._converged = False
    
    def on_epoch_end(self, epoch, logs=None):
        if not self._converged:
            self._print('\n---- refreshing iterates ----')

            start = time()

            self.model.refresh_iterates(
                convergence_threshold=self._short_convergence_threshold,
                batch_size=self._batch_size,
                verbose=self._verbose)

            self._print(f'   > done: {(time() - start):.2f} seconds')

    def on_train_end(self, logs=None):
        self._print('---- refreshing iterates precisely ----')

        start = time()
        self.model.refresh_iterates(
            convergence_threshold=self._convergence_threshold,
            batch_size=self._batch_size,
            verbose=self._verbose)

        self._print(f'   > done: {(time() - start):.2f} seconds')



import numpy as np
import tensorflow as tf

from tensorflow.keras.optimizers import Adam 
from tensorflow.keras.utils import Progbar

from gloro.constants import EPS


def lipschitz_lb(f, X1, X2, iterations=1000, verbose=True):

    optimizer = Adam(lr=0.0001)

    X1 = tf.Variable(X1, name='x1', dtype='float32')
    X2 = tf.Variable(X2, name='x2', dtype='float32')
    
    max_L = None

    if verbose:
        pb = Progbar(iterations, stateful_metrics=['LC'])
    
    for _ in range(iterations):
        with tf.GradientTape() as tape:
            y1 = f(X1)
            y2 = f(X2)
            
            # The definition of the margin is not entirely symmetric: the top
            # class must remain the same when measuring both points. We assume
            # X1 is the reference point for determining the top class.
            original_predictions = tf.cast(
                tf.equal(y1, tf.reduce_max(y1, axis=1, keepdims=True)), 
                'float32')
            
            # This takes the logit at the top class for both X1 and X2.
            y1_j = tf.reduce_sum(
                y1 * original_predictions, axis=1, keepdims=True)
            y2_j = tf.reduce_sum(
                y2 * original_predictions, axis=1, keepdims=True)
            
            margin1 = y1_j - y1
            margin2 = y2_j - y2

            axes = tuple((tf.range(len(X1.shape) - 1) + 1).numpy())
            
            L = tf.abs(margin1 - margin2) / (tf.sqrt(
                tf.reduce_sum((X1 - X2)**2, axis=axes)) + EPS)[:,None]

            loss = -tf.reduce_max(L, axis=1)
            
        grad = tape.gradient(loss, [X1, X2])

        optimizer.apply_gradients(zip(grad, [X1, X2]))
        
        if max_L is None:
            max_L = L
        else:
            max_L = tf.maximum(max_L, L)

        if verbose:
            pb.add(1, [('LC', tf.reduce_max(max_L))])
        
    return tf.reduce_max(max_L)


def local_lipschitz_lb(f, X1, X2, eps, iterations=1000, verbose=True):

    optimizer = Adam(lr=0.0001)

    X0 = tf.constant(X1, dtype='float32')
    y0 = f(X0)

    X1 = tf.Variable(X1, name='x1', dtype='float32')
    X2 = tf.Variable(X2, name='x2', dtype='float32')
    
    max_L = None

    if verbose:
        pb = Progbar(iterations, stateful_metrics=['max_LC', 'mean_LC'])
    
    for i in range(iterations):
        with tf.GradientTape() as tape:

            axes = tuple((tf.range(len(X1.shape) - 1) + 1).numpy())

            # Project so that X1 and X2 are at distance at most `eps` from X0.
            delta1 = X1 - X0
            dist1 = tf.sqrt(tf.reduce_sum(
                delta1 * delta1, axis=axes, keepdims=True))
            
            delta2 = X2 - X0
            dist2 = tf.sqrt(tf.reduce_sum(
                delta2 * delta2, axis=axes, keepdims=True))
            
            # Only project if `dist` > `eps`.
            where_dist_gt_eps1 = tf.cast(dist1 > eps, 'float32')
            where_dist_gt_eps2 = tf.cast(dist2 > eps, 'float32')

            X1.assign(
                (X0 + eps * delta1 / (dist1 + EPS)) * where_dist_gt_eps1 + 
                X1 * (1 - where_dist_gt_eps1))
            X2.assign(
                (X0 + eps * delta2 / (dist2 + EPS)) * where_dist_gt_eps2 + 
                X2 * (1 - where_dist_gt_eps2))

            y1 = f(X1)
            y2 = f(X2)
            
            # The definition of the margin is not entirely symmetric: the top
            # class must remain the same when measuring both points. We assume
            # X0 is the reference point for determining the top class.
            original_predictions = tf.cast(
                tf.equal(y0, tf.reduce_max(y0, axis=1, keepdims=True)), 
                'float32')
            
            # This takes the logit at the top class for both X1 and X2.
            y1_j = tf.reduce_sum(
                y1 * original_predictions, axis=1, keepdims=True)
            y2_j = tf.reduce_sum(
                y2 * original_predictions, axis=1, keepdims=True)
            
            margin1 = y1_j - y1
            margin2 = y2_j - y2

            L = tf.abs(margin1 - margin2) / (tf.sqrt(
                tf.reduce_sum((X1 - X2)**2, axis=axes)) + EPS)[:,None]

            loss = -tf.reduce_max(L, axis=1)
           
        if i < iterations - 1: 
            grad = tape.gradient(loss, [X1, X2])

            optimizer.apply_gradients(zip(grad, [X1, X2]))
        
        if max_L is None:
            max_L = L
        else:
            max_L = tf.maximum(max_L, L)

        if verbose:
            metrics = [
                ('max_LC', tf.reduce_max(max_L)), 
                ('mean_LC', tf.reduce_mean(max_L))
            ]
            
            pb.add(1, metrics)
        
    return tf.reduce_max(max_L), tf.reduce_mean(max_L)



import os

from setuptools import setup, find_packages


with open('README.md', 'r', encoding='utf-8') as f:
    long_description = f.read()

def read(rel_path):
    here = os.path.abspath(os.path.dirname(__file__))
    # intentionally *not* adding an encoding option to open, See:
    #   https://github.com/pypa/virtualenv/issues/201#issuecomment-3145690
    with open(os.path.join(here, rel_path), 'r') as fp:
        return fp.read()

def get_version(rel_path):
    for line in read(rel_path).splitlines():
        if line.startswith('__version__'):
            return line.split("'")[1]

    raise RuntimeError('Unable to find version string.')

with open('requirements.txt', 'r') as requirements:
    setup(
        name='gloro',
        version=get_version('gloro/__init__.py'),
        install_requires=list(requirements.read().splitlines()),
        packages=find_packages(),
        description=
            'library for training globally-robust neural networks',
        python_requires='>=3.6',
        author='Klas Leino',
        author_email='kleino@cs.cmu.edu',
        classifiers=[
            'Programming Language :: Python :: 3',
            'License :: OSI Approved :: MIT License',
            'Operating System :: OS Independent'],
        long_description=long_description,
        long_description_content_type='text/markdown',
        setup_requires=['setuptools_scm'],
    )
